#include <M5Stack.h>
#include "ESP32_SD_UTF8toSJIS.h"
#include "ESP32_SD_ShinonomeFNT.h"
#include <WiFi.h>

// POSTで読み取るデータサイズを定義
#define BUFSIZE 256
// BOARDに表示可能な文字数を定義
#define BOARDWIDTH 100
// BOARDに表示するデコード済みのテキストを格納
char* decodeWordValue;
char* decodeColorValue;

WiFiServer server(80);

#define SSID "**********"   // 無線LANアクセスポイントのSSID
#define PASS "**********"       // パスワード

String lip;
#define TFT_GRAY 0xC618
#define TFT_DARK 0x7BEF
const char* UTF8SJIS_file = "/font/Utf8Sjis.tbl"; //UTF8 Shift_JIS 変換テーブルファイル名を記載しておく
const char* Shino_Zen_Font_file = "/font/shnmk16.bdf"; //全角フォントファイル名を定義
const char* Shino_Half_Font_file = "/font/shnm8x16.bdf"; //半角フォントファイル名を定義
const uint8_t CS_SD = 4; //SD card CS ( Chip Select )
const uint8_t panelW=64;
const uint8_t panelH=16;
const uint8_t dTime=10;
const uint8_t msgL=4;
const String SPACE="　　　　";
String msg[10]={SPACE};
String xmsg[10]={SPACE};
String postMsg,postColor,dcols,mcols;
char mbuf[100];
uint16_t tcolor[8]={TFT_BLACK,TFT_WHITE,TFT_RED,TFT_GREEN,TFT_CYAN,TFT_MAGENTA,TFT_YELLOW,TFT_BLUE};
uint8_t col[40]={7};
uint8_t xcol[40]={7};
uint8_t dcol[40]={7};
uint8_t mcol[40]={7}; 
uint8_t rcol[40]={7};
uint8_t msgN,xmsgN,demoN,menuN,dcolN,colN,rcolN;
uint16_t sj_length; //Shift_JISコードの長さ
uint8_t xo=32;
uint8_t yo=80;
uint8_t font_buf[32][16]; //東雲フォントデータ
uint8_t buf[16]={0}; 
uint8_t dot[128][16]={0};
uint8_t ptr=0;
uint8_t msgPtr=0;
uint8_t r=1;
uint8_t cn=0;
uint8_t cnt=0;
uint8_t LBmode=0;
uint8_t SDflag=0;
uint8_t Cflag=0;
uint16_t x,y,i,cl,xb,len;

ESP32_SD_ShinonomeFNT SFR(CS_SD, 24000000);

//server.begin();

void setup() {
  M5.begin();
  delay(500);
  M5.Lcd.fillScreen(TFT_DARK);
  M5.Lcd.setTextSize(2);
  M5.Lcd.setTextColor(TFT_YELLOW);
  M5.Lcd.setCursor(30,190); M5.Lcd.print("SELECT BUTTON");
  M5.Lcd.drawRect(15,68,290,95,TFT_GRAY);
  M5.Lcd.drawRect(11,64,298,103,TFT_GRAY);
  M5.Lcd.fillRect(16,69,289,91,TFT_BLACK);
  
  SFR.SD_Shinonome_Init3F(UTF8SJIS_file, Shino_Half_Font_file, Shino_Zen_Font_file);
  SDflag=1;
  xmsgN=0;
  menuN=7;
  Cflag=0;
  String menu[7]={"Ａボタン","（デモ）","かＢボタ", "ン（通信","）を押し","て下さい","。　　　"};
  mcols="3";
  msgN=menuN;msgPtr=0;
  for (i=0; i<msgN; i++) msg[i]=menu[i];
  for (i=0; i<msgN*msgL; i++) col[i]=mcols.toInt();
  ptr=0;msgPtr=0; cnt=0;
  readMessage(0);
  readMessage(64);
}

void loop() {
  if (M5.BtnA.wasPressed()){
    demoMode();
  } else if(M5.BtnB.wasPressed()){
    commMode();
  } else if(M5.BtnC.wasPressed()){
    closeMode();
  }
  M5.update();
  if (LBmode <3){
    if (LBmode==2 && Cflag==1) getPOST();
    dispMessage(ptr); delay(dTime);
    ptr++; if (ptr==128) ptr=0;
    if (ptr==0 || ptr==64) readMessage(64-ptr);
  }
}

void readMessage(uint8_t p){
  i=0;
  sj_length = SFR.StrDirect_ShinoFNT_readALL(msg[msgPtr], font_buf);
  for (i=0; i<msgL;i++){
    for(y=0; y<panelH; y++){
      for (x=0; x<8; x++){
        dot[x+p+i*16][y]=bitRead(font_buf[2*i][y],7-x)*10+col[cnt];            
      }
      for (x=8; x<16; x++){
        dot[x+p+i*16][y]=bitRead(font_buf[2*i+1][y],15-x)*10+col[cnt];
      } 
    }
    cnt++; if(cnt==msgN*msgL) cnt=0;  
  }
  msgPtr++; if(msgPtr==msgN) msgPtr=0;
}


void demoMode(){
  LBmode=1;
  if (SDflag==0) {
    SFR.SD_Shinonome_Init3F(UTF8SJIS_file, Shino_Half_Font_file, Shino_Zen_Font_file);
    SDflag=1;
  }
  demoN=5;
  String demo[5]= {"ＡＤＣＬ","カラーＬ","ＥＤ電光","掲示板で","す！！　"};
  dcols="123456";
  dcolN=dcols.length();
  for (i=0; i<dcolN; i++) {
    String d1=dcols.substring(i,i+1);
    dcol[i]=d1.toInt();
  }
  cn=0;
  for (i=0; i<demoN*msgL; i++) {
    col[i]=dcol[cn];
    cn++; if (cn==dcolN) cn=0;
  }
  msgN=demoN; ptr=0; cnt=0; msgPtr=0;
  for (i=0; i<msgN; i++) msg[i]=demo[i];
  M5.Lcd.fillRect(16,69,289,91,TFT_BLACK);
  M5.Lcd.setTextColor(TFT_YELLOW);
  M5.Lcd.fillRect(30,190,290,20,TFT_DARK);
  M5.Lcd.setCursor(30,190); M5.Lcd.print("DEMONSTRATION");
  ptr=0;msgPtr=0;
  readMessage(0);
  readMessage(64);
}

void commMode(){
  LBmode=2;
  M5.Lcd.fillRect(30,190,300,20,TFT_DARK);
  M5.Lcd.setTextColor(TFT_GREEN);
  M5.Lcd.setCursor(30,190); M5.Lcd.print("COMMUNICATION"); 
  M5.Lcd.fillRect(16,69,289,91,TFT_BLACK);
  if(Cflag==0){
    int waiting=0;
    M5.Lcd.setTextColor(TFT_YELLOW);
    M5.Lcd.fillRect(30,25,210,20,TFT_DARK);
    M5.Lcd.setCursor(50,25); M5.Lcd.print("Connecting");
    WiFi.mode(WIFI_STA);                    // 無線LANをSTAモードに設定
    WiFi.begin(SSID,PASS);                  // 無線LANアクセスポイントへ接続
    while(WiFi.status() != WL_CONNECTED){   // 接続に成功するまで待つ
      delay(100);                           // 待ち時間処理
      waiting++;
      if (waiting % 10==0) {
        M5.Lcd.setCursor(160+waiting,25); M5.Lcd.print(".");
      }
      if (waiting==100) break;
    }
    if (WiFi.status()==WL_CONNECTED){
      lip=WiFi.localIP().toString();  
      // IPアドレスをLCDに表示
      M5.Lcd.fillRect(30,25,210,20,TFT_DARK);
      M5.Lcd.setTextColor(TFT_YELLOW);
      M5.Lcd.setCursor(50,25); M5.Lcd.print("MY IP:"+lip);
      Cflag=1;
    }
  }
  if (xmsgN==0){
    if(SDflag==0) {
      SFR.SD_Shinonome_Init3F(UTF8SJIS_file, Shino_Half_Font_file, Shino_Zen_Font_file);
      SDflag=1;
    }
    String msg1[2]= {"受信待ち","です　　"};
    msgN=2; ptr=0; cnt=0; msgPtr=0;
    for (i=0; i<msgN; i++) msg[i]=msg1[i];
    for (i=0; i<msgN*msgL; i++) col[i]=2;
    ptr=0; msgPtr=0; cnt=0;
    readMessage(0);
    readMessage(64);
  } else {
    updateMessage();
  }
}

void closeMode(){
  LBmode=3;
  if (SDflag==1) {
    SFR.SD_Shinonome_Close3F(); //StrDirect_ShinoFNT_readALL関数を使う場合、必ず閉じておく
    SDflag=0;
  }
  M5.Lcd.fillRect(16,69,289,91,TFT_BLACK);
  M5.Lcd.fillRect(30,190,290,20,TFT_DARK);
  M5.Lcd.setTextColor(TFT_WHITE);
  M5.Lcd.setCursor(30,190); M5.Lcd.print("SD CLOSED"); 
  SDflag=0;
}

void dispMessage(uint8_t p){
  i=0;
  for (x=p; x<p+64; x++){
    xb=x; if (x>127) xb=x-128;
    for (y=0; y<panelH; y++){
      cl=dot[xb][y] % 10;
      if (dot[xb][y]>=10) M5.Lcd.fillCircle(xo+i*4,yo+y*4,r,tcolor[cl]);
      else M5.Lcd.fillCircle(xo+i*4,yo+y*4,r,TFT_BLACK);
    }
    i++; 
  }
  delay(dTime);
}

void recvMessage(){
  postMsg=postMsg+"　　　　";
  int n=zenLength(postMsg);
  xmsgN=n/4;
  if (xmsgN>10) xmsgN=10;
  for (i=0; i<xmsgN; i++){
    xmsg[i]=postMsg.substring(i*12,i*12+12);
  }
  rcolN=postColor.length();
  for (i=0; i<rcolN; i++){
    String c1=postColor.substring(i,i+1);
    rcol[i]=c1.toInt();
  }
  int cn=0;
  for (i=0; i < xmsgN*msgL; i++){
    xcol[i]=rcol[cn];
    cn++;
    if (cn==rcolN) cn=0;
  }
}

void updateMessage(){
  if (xmsgN>0){
    msgN=xmsgN;
    for (i=0; i<msgN; i++) msg[i]=xmsg[i];
    for (i=0; i<msgN*msgL; i++) col[i]=xcol[i];             
    ptr=0; msgPtr=0; cnt=0;
    msgPtr=0; cnt=0;
    dot[128][16]={0};
    readMessage(0);
    readMessage(64);
  }    
}

void getPOST(){
  server.begin();
  // クライアントオブジェクトを取得
  WiFiClient client = server.available();
  // クライアントが存在するとき、clientはtrue
  if (client) {
    // クライアントからPOSTメソッドで呼ばれたらtrue
    boolean isPOST = false;
    // リクエストヘッダに改行（\r）以外が含まれたらfalse
    boolean currentLineIsBlank = true;
    // リクエストヘッダの終了（\r\n\r\n）を検出したらtrue
    boolean reqHeaderIsEnd = false;
    // リクエストデータを格納
    // 最大サイズはBUFSIZEで規定
    char reqParam[BUFSIZE];
    // リクエストデータを読むときに使用するインデックス
    long idx = 0;
    // POSTで送信されたテキストを格納
    char* wordValue;
    char* colorValue;
    // クライアントが存在する間、ぐるぐるまわる
    while (client.connected()) {
      // クライアントから受信するデータが存在するときはtrueを返す
      if (client.available()) {
        char c = client.read();
        // （\r）\n（\r）\nのパターンがあらわれた場合、リクエストヘッダの終わりと見なしてreqHeaderIsEndをtrueにする
        if (c == '\n' && currentLineIsBlank) {
          reqHeaderIsEnd = true;
        }
        if (c == '\n') {
          currentLineIsBlank = true;

          // 行末にきたらidxをリセット
          idx = 0;
          // 行頭が"POST "ではじまっていたら、isPOSTをtrueにする
          if (strncmp(reqParam, "POST ", 5) == 0) {
            isPOST = true;
          }
        } 
        else if (c != '\r') {
          currentLineIsBlank = false;
          // リクエストデータ1行分をreqParamに格納する
          // 1行のデータサイズがPOSTデータの規定最大サイズBUFSIZE-1（-1は'\0'付加を考慮した分）を超える場合は、データを破棄
          if (idx < BUFSIZE - 1) {
            reqParam[idx] = c;
            idx++;
          }
        }
      }
      // クライアントから受信するデータがなくなったらループを抜ける
      else {
        break;
      }
    }
    // POSTで呼び出された場合のデータ解析とデコード処理
    if (isPOST) {
      // POSTデータを'&'で分割したパラメタを格納
      char* splitReqParam;
      // POSTデータをstrtok_rでパラメタに分割するときに使用
      char* reqParamPtr;
      // パラメタを'='で分割した左辺（パラメタ名）を格納
      char* paramName;
      // パラメタを'='で分割した右辺（パラメタ値）を格納
      char* paramValue;
      // パラメタをstrtok_rで名前と値に分割するときに使用
      char* nameValuePtr;
      // reqParamに格納した文字列の後ろにNULLを追加して、文字列の終わりを示す
      reqParam[idx] = '\0';

      // POSTデータを'&'で分割してsplitReqParamに格納
      for (splitReqParam = strtok_r(reqParam, "&", &reqParamPtr); splitReqParam != NULL; splitReqParam = strtok_r(NULL, "&", &reqParamPtr)) {
        // パラメタsplitReqParamを'='で分割して、名前をparamName、値をparamValueに格納
        // ネストしたいので、strtokの代わりにstrtok_rを使用
        char* paramName = strtok_r(splitReqParam, "=", &nameValuePtr);
        char* paramValue = strtok_r(NULL, "=", &nameValuePtr);

        // パラメタ名が"word"であれば、テキストとみなしてwordValueに値を格納
        if (strcmp(paramName, "word") == 0) {
          wordValue = paramValue;
        }
        // パラメタ名が"color"であれば、テキストとみなしてwordValueに値を格納
        if (strcmp(paramName, "color") == 0) {
          colorValue = paramValue;
        }
      }

      // POSTで送信されたデータはパーセントエンコーディングされているため、デコードする関数を呼ぶ
      // デコードされた文字列はdecodeWordValueに格納
      decodeWordValue = strdup(wordValue);
      urlDecode(wordValue, decodeWordValue);
      // POSTで送信されたデータはパーセントエンコーディングされているため、デコードする関数を呼ぶ
      // デコードされた文字列はdecodeWordValueに格納
      decodeColorValue = strdup(colorValue);
      urlDecode(colorValue, decodeColorValue);
    }
    // クライアントにデータを返す
    client.println("HTTP/1.1 200 OK");
    client.println("Content-Type: text/html");
    client.println("Connnection: close");
    client.println();
    client.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    client.println("<!DOCTYPE html>");
    client.println("<html lang=\"ja\">");
    client.println("<head>");
    client.println("<meta http-equiv=\"content-type\" content=\"text/html\" />");
    client.println("<title>");
    client.println("電光掲示板");
    client.println("</title>");
    client.println("</head>");
    client.println("<body>");
    client.println("<font size='3'>");
    client.println("電光掲示板<br><br>");
    client.println("<form method=\"post\">");
    client.println("送信文字列:");
    client.println("<input type=\"text\" name=\"word\" value='' size=40"); if (isPOST) { client.print(decodeWordValue); } client.print("\" />");
    client.println("<br>");
    client.println("文字色設定:");
    client.println("<input type=\"text\" name=\"color\" value='' size=40"); if (isPOST) { client.print(decodeColorValue); } client.print("\" />");
    client.println("<br>");
    client.println("<input type=\"submit\" value=\"SEND\" />");  
    client.println("<input type=\"reset\" value=\"CLEAR\" />");
    client.println("</form>");
    client.println("</font>");
    client.println("</body>");
    client.println("</html>");
    // クライアントをサーバから切断
    delay(1);
    client.stop();
    // POSTで呼び出された場合
    if (isPOST) {  
      // dispMsgはBOARDに表示する文字列を格納する変数
      // 変数のサイズはBOARDに表示可能な文字数+1（'\0'を考慮して1文字分多く確保）
      char dispMsg1[BOARDWIDTH + 1];
      char dispMsg2[BOARDWIDTH + 1];
      // テキストからBOARDに表示可能な文字数を切り取ってdispMsgに格納
      strncpy(dispMsg1, decodeWordValue, BOARDWIDTH);
      // strncpyは'\0'をつけてくれないので、自分でつける
      dispMsg1[BOARDWIDTH] = '\0';
      strncpy(dispMsg2, decodeColorValue, BOARDWIDTH);
      // strncpyは'\0'をつけてくれないので、自分でつける
      dispMsg2[BOARDWIDTH] = '\0';
      postMsg=convToZen(dispMsg1);
      postColor=dispMsg2;
      recvMessage();
      updateMessage();
    }
  }
}

boolean urlDecode(char* str, char* dstStr) {
  // POSTで読み取ったデータはパーセントエンコーディングされているため、デコードを実施
  // char* str : 変換する文字列を格納、 char* dstStr : 変換後の文字列を格納
  if (str != NULL) {
    long dstIdx = 0;

    for(long idx = 0; idx < strlen(str); idx++) {
      if (str[idx] == '%') {
        char tmpHex[] = { str[idx+1], str[idx+2] };
        dstStr[dstIdx] = strtol(tmpHex, NULL, 16);
        idx += 2;
      }
      else if (str[idx] == '+') {
        dstStr[dstIdx] = ' ';
      }
      else {
        dstStr[dstIdx] = str[idx];
      }
      dstIdx++;
    }
    dstStr[dstIdx] = '\0';
  }
  return true;
}

int zenLength(String s){  //全角文字列の長さを求める
  return s.length()/3;
}

String convToZen (String s1){ // 半角英数文字を全角英数文字に変換する
  char chr1[100]={0};
  char chr2[100]={0};
  int ln=s1.length();
  char c;
  s1.toCharArray(chr1,ln+1);
  int ptr1=0;
  int ptr2=0;
  String s2;
  
  while (ptr1<ln){
    c=chr1[ptr1];
    if (c<0x60){
      chr2[ptr2]=0xEF;
      chr2[ptr2+1]=0xBC;
      chr2[ptr2+2]=c+0x60;
      ptr1++;
      ptr2=ptr2+3;
    } else if (c<0x80){
      chr2[ptr2]=0xEF;
      chr2[ptr2+1]=0xBD;
      chr2[ptr2+2]=c+0x20;
      ptr1++;
      ptr2=ptr2+3;      
    } else {
      chr2[ptr2]=chr1[ptr1];
      chr2[ptr2+1]=chr1[ptr1+1];
      chr2[ptr2+2]=chr1[ptr1+2];
      ptr1=ptr1+3;
      ptr2=ptr2+3;
    }
  }
  s2=chr2;
  return s2;
}
